// ------------------------------- //
// -------- Start of File -------- //
// ------------------------------- //
// ----------------------------------------------------------- //
// C++ Header File Name: devcache.h
// Compiler Used: MSVC40, egcs-2.90.29, HP CPP 10.24
// Produced By: Doug Gaer   
// File Creation Date: 09/20/1999
// Date Last Modified: 07/17/2000
// Copyright (c) 2000 Douglas M. Gaer
// ----------------------------------------------------------- // 
// ---------- Include File Description and Details  ---------- // 
// ----------------------------------------------------------- // 
/*
The VBS C++ classes are copyright (c) 2000, by Douglas M. Gaer.
All those who put this code or its derivatives in a commercial
product MUST mention this copyright in their documentation for
users of the products in which this code or its derivative
classes are used. Otherwise, you have the freedom to redistribute
verbatim copies of this source code, adapt it to your specific
needs, or improve the code and release your improvements to the
public provided that the modified files carry prominent notices
stating that you changed the files and the date of any change.

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.
THE ENTIRE RISK OF THE QUALITY AND PERFORMANCE OF THIS SOFTWARE
IS WITH YOU. SHOULD ANY ELEMENT OF THIS SOFTWARE PROVE DEFECTIVE,
YOU WILL ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR, OR
CORRECTION.

The DeviceCache class is an abstract base class used to define
the methods needed to interface a device with memory cache.
The derived class determines the type of device the cache will
be linked to: a file, TCP/IP socket, serial port, etc. The derived
class is also responsible for defining the methods used to read
and write to and from the device as well as keep track of the
device status. The device cache will buffer a stream of binary
data as is it is written to or read from the device the derived
class is using.
*/
// ----------------------------------------------------------- //   
#ifndef __DEVICE_CACHE_HPP__
#define __DEVICE_CACHE_HPP__

#include <stddef.h>
#include "devtypes.h"

// Forward class declarations
class vbDeviceBucketb;
class vbDeviceBucketCacheb;
class vbDeviceCachePtrb;

// General definition of the bucket data type.
// Each data type is set by a conditional directive that allows the
// device cache classes to be used for several data types. Each type
// must be determined at compile time and only one type can be used
// for each compile. Variable length cache buckets will be used by
// default
#if defined(__SMEMORY_BLOCK__) // Fixed length memory (static) blocks
#include "memblock.h"
typedef MemoryBlock MBTYPE; 

#elif defined(__VMEMORY_BUFFER__) // Variable length memory buffers
#include "membuf.h"
typedef MemoryBuffer MBTYPE; 

#else // Default to variable length memory buffers
#include "membuf.h"
typedef MemoryBuffer MBTYPE; 
#endif

class vbDeviceCache
{
public:
  vbDeviceCache() { }
  virtual ~vbDeviceCache();
  
public: // Derived class interface
  virtual void Read(void *buf, unsigned Bytes, vbDeviceTypes dev) = 0;
  virtual void Write(const void *buf, unsigned Bytes, vbDeviceTypes dev) = 0;

public:
  int ReadyForWriting() { return ready_for_writing; }
  int ReadyForReading() { return ready_for_reading; }
  int IsOK() { return is_ok; }
  
protected: // Device flags set by the derived class
  int is_ok; 
  int ready_for_writing;
  int ready_for_reading;
};

class vbDeviceBucketb
// The vbDeviceBucketb class is abstract base class used to defined
// how each bucket will be loaded/stored and which device will
// be used for every input/output operation. vbDeviceBucketb objects
// are organized into a doubly linked list by the vbDeviceBucketCacheb
// base class.
{
public:
  vbDeviceBucketb() { } // Other classes initialize buckets

public:
  void Lock() { ++refcnt; }
  void Unlock() { --refcnt; }
  int IsLocked() { return refcnt; }
  int IsNull() { return o_device == vbDEVICE_VOID; }
  int IsDirty() { return Dirty == 1; }
  void Release() { refcnt = 0; }
  void SetDirty() { Dirty = 1; }
  void SetDirty(vbDeviceCache &dev) { Dirty = 1; if(o_device) Store(dev); } 
  void Flush(vbDeviceCache &dev) { if(o_device && Dirty) Store(dev); } 
  virtual void Fetch(vbDeviceCache &dev, unsigned bytes) = 0; 
  virtual void Store(vbDeviceCache &dev) = 0; 
  int GetRefCnt() { return refcnt; }
  char GetDirty() { return Dirty; }
  vbDeviceBucketb *GetPrior() { return Prior; }
  vbDeviceBucketb *GetNext() { return Next; }

public:
  friend class vbDeviceBucketCacheb;
  friend class vbDeviceCachePtrb;

protected:
  void MakeNull() {
    refcnt = 0; Dirty = 0; o_device = i_device = vbDEVICE_VOID;
  }

protected:
  vbDeviceTypes o_device; // Output device this bucket is connected to
  vbDeviceTypes i_device; // Input device this bucket is connected to
  int refcnt;             // How many cache pointers pointing to this bucket
  char Dirty;             // True when the bucket is in use
  vbDeviceBucketb *Prior; // Pointer to adjacent buckets in the cache
  vbDeviceBucketb *Next;  // Most recently acquired bucket
};

class vbDeviceBucketCacheb
// The vbDeviceBucketCacheb class is used as a base class for the
// vbDeviceBucketCache class. It incorporates functions that are
// independent of the bucket type. The cache is used to handle requests
// for memory-based objects. The cache must determine whether an object
// is already loaded. If the object is not loaded the cache reserves a
// bucket and loads the object into memory. This cache design uses cache
// pointers to reference cache buckets directly, with each cache pointer
// being initialized after the bucket is reserved. 
{
protected:
  vbDeviceBucketCacheb(vbDeviceBucketb *b, int n, unsigned bkt_size);
  virtual ~vbDeviceBucketCacheb();

protected:
  vbDeviceBucketb *AcquireBkt();
  void MoveToFront(vbDeviceBucketb *b);

public:
  friend class vbDeviceCachePtrb;
  void Connect(vbDeviceCache *dev) { if(devptr) Clear(); devptr = dev; } 
  void Disconnect() { if(devptr) Clear(); devptr = 0; }
  int Flush(int empty_bkts = 0);
  int rFlush(int empty_bkts = 0);
  void Clear();
  vbDeviceBucketb *ReserveBkt(vbDeviceTypes o_dev,
			      vbDeviceTypes i_dev,
			      int ensure_loaded, unsigned bytes);
  vbDeviceBucketb *GetHead() { return head; } 
  int GetBuckets() { return nbuckets; }
  unsigned BucketsInUse();

protected:
  int nbuckets;          // Size of cache
  vbDeviceCache *devptr; // Device cache is connected to
  vbDeviceBucketb *head; // Most recently reserved bucket
  vbDeviceBucketb *tail; // Least recently reserved bucket
};

class vbDeviceCachePtrb
// The vbDeviceCachePtrb class is used as a data independent base class for
// the vbDeviceCachePtr class. Cache pointers are used to work in conjunction
// with the reference counted cache buckets. Each cache pointer stores
// a pointer to the bucket containing the memory buffer of the object
// and a pointer to the cache the cache pointer is connected to.
{
public:
  ~vbDeviceCachePtrb() { Release(); }

protected:
  vbDeviceCachePtrb(vbDeviceBucketCacheb &c,
		    vbDeviceTypes o_dev, vbDeviceTypes i_dev);
  vbDeviceCachePtrb(const vbDeviceCachePtrb &c);

private:
  void operator=(const vbDeviceCachePtrb &c) { }; // Disallowed

public:
  void Grab() { if (!bound) Bind(); }
  void Release();
  void SetOutputDevice(vbDeviceTypes dev) { o_device = dev; }
  void SetInputDevice(vbDeviceTypes dev) { i_device = dev; }
  vbDeviceTypes GetOutputDevice() { return o_device; }
  vbDeviceTypes GetInputDevice() { return i_device; }
  
protected:
  vbDeviceBucketb *AllocBucket(vbDeviceTypes o_dev, vbDeviceTypes i_dev);
  vbDeviceBucketb *AllocBucket(vbDeviceTypes o_dev, vbDeviceTypes i_dev,
			       unsigned buffer_size);
  void Copy(const vbDeviceCachePtrb &c);
  int Bind();

protected:
  vbDeviceTypes o_device;      // Output device this bucket is connected to
  vbDeviceTypes i_device;      // Input device this bucket is connected to
  vbDeviceBucketb *bkt;        // Pointer to bucket in the cache
  vbDeviceBucketCacheb *cache; // Pointer to the cache being used
  char bound;                  // True if pointer is bound to a bucket
};

class vbDeviceBucket : public vbDeviceBucketb, public MBTYPE
// The cache bucket class is derived from the vbDeviceBucketb class,
// creating a family of classes specific to the data type stored.
{
public:
  vbDeviceBucket() { } 

public:
  void Fetch(vbDeviceCache &dev, unsigned bytes);
  void Store(vbDeviceCache &dev);
};

class vbDeviceBucketCache : public vbDeviceBucketCacheb
// The vbDeviceBucketCache class is derived from the vbDeviceBucketCacheb
// class. It adds two functions to handle the allocation and de-allocation
// of buckets.
{
public:
  vbDeviceBucketCache(int n);
  virtual ~vbDeviceBucketCache();

public:
  vbDeviceBucket *buckets;
};

class vbDeviceCachePtr : public vbDeviceCachePtrb
// Cache pointers are used to work in conjunction with the
// reference counted cache buckets. Each cache pointer stores 
// a pointer to the bucket containing the memory buffer of the
// object and a pointer to the cache the cache pointer is connected
// to.
{
public:
  vbDeviceCachePtr(vbDeviceBucketCache &c, vbDeviceTypes o_dev,
		   vbDeviceTypes i_dev)
    : vbDeviceCachePtrb(c, o_dev, i_dev) { }
  vbDeviceCachePtr(const vbDeviceCachePtr &c) : vbDeviceCachePtrb(c) { }
  void operator=(const vbDeviceCachePtr &c) { Copy(c); }

public:
  vbDeviceBucket &operator*() {
    if(!bound) Bind(); return *(vbDeviceBucket *)bkt; }

  vbDeviceBucket *operator->() {
    if (!bound) Bind(); return (vbDeviceBucket *)bkt; }

  MBTYPE *Alloc() 
  // Reserves a cache bucket for this cache pointer without loading
  // any data in the cache bucket. The calling function must load
  // the bucket after this function reserves the bucket.
  { 
    return (MBTYPE *)((vbDeviceBucket *)(AllocBucket(this->o_device,
						     this->i_device))); 
  }

  MBTYPE *Alloc(unsigned bytes) 
  // Overload used to automatically load a specified number of bytes
  // into the bucket from the input device the bucket is connected to.
  {
    return (MBTYPE *)((vbDeviceBucket *)(AllocBucket(this->o_device,
						     this->i_device, bytes))); 
  }
};

#endif // __DEVICE_CACHE_HPP__
// ----------------------------------------------------------- // 
// ------------------------------- //
// --------- End of File --------- //
// ------------------------------- //
