// ------------------------------- //
// -------- Start of File -------- //
// ------------------------------- //
// ----------------------------------------------------------- //
// C++ Header File Name: membuf.h 
// Compiler Used: MSVC40, HP CPP 10.24, egcs-2.90.29
// Produced By: Doug Gaer 
// File Creation Date: 09/20/1999
// Date Last Modified: 07/17/2000
// Copyright (c) 2000 Douglas M. Gaer
// ----------------------------------------------------------- // 
// ---------- Include File Description and Details  ---------- // 
// ----------------------------------------------------------- // 
/*
The VBS C++ classes are copyright (c) 2000, by Douglas M. Gaer.
All those who put this code or its derivatives in a commercial
product MUST mention this copyright in their documentation for
users of the products in which this code or its derivative
classes are used. Otherwise, you have the freedom to redistribute
verbatim copies of this source code, adapt it to your specific
needs, or improve the code and release your improvements to the
public provided that the modified files carry prominent notices
stating that you changed the files and the date of any change.

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.
THE ENTIRE RISK OF THE QUALITY AND PERFORMANCE OF THIS SOFTWARE
IS WITH YOU. SHOULD ANY ELEMENT OF THIS SOFTWARE PROVE DEFECTIVE,
YOU WILL ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR, OR
CORRECTION.

The MemoryBuffer class is used to create variable length memory
buffers. 
*/
// ----------------------------------------------------------- //   
#ifndef __MEMBUF_HPP__
#define __MEMBUF_HPP__

#include <string.h>

// Memory representation object
class MemoryRep 
{
private:
  friend class MemoryBuffer;
  MemoryRep() { RefCount = 1; }

private:
  void *operator new(size_t RepSize, unsigned Bytes = 0);
  void operator delete(void *ptr);
  
private:
  unsigned int RefCount;     // Reference count for memory reps
  unsigned char Data[1];     // Start of data
  static MemoryRep Null_Ptr; // For null objects
};

// Memory buffer class
class MemoryBuffer
{
public:
  enum { DefSize = 16, DefInc = 16, NoMatch = -1 };
  MemoryBuffer(unsigned Bytes = DefSize, int GB = DefInc) { Alloc(Bytes, GB); }
  MemoryBuffer(const void *buf, unsigned Bytes, int GB = DefInc);
  ~MemoryBuffer();
  MemoryBuffer(const MemoryBuffer &buf);
  MemoryBuffer &operator=(const MemoryBuffer &buf);

public:
  unsigned length() const { return Len; }       // Returns logical length
  int resize(unsigned bytes, int keep = 0);     // Adjust the logical length 
  unsigned char *m_buf() { return BufferPtr; }  // Return buffer's pointer 
  const unsigned char *m_buf() const { return BufferPtr; } 
  void Load(const void *buf, unsigned Bytes);   // Load the buffer
  void Clear();                                 // Clear the buffer

  // Append specified number of bytes from buf onto end of this buffer
  void Cat(const void *buf, unsigned Bytes) { InsReplAt(Len, buf, Bytes); }
  void Cat(unsigned char byte) { Cat((void *)&byte, 1); }
  void Cat(char byte) { Cat((void *)&byte, 1); }

  // Copy data from buf into this object
  void Copy(const MemoryBuffer &buf); 

  // Search functions
  int Find(void *buf, unsigned Offset = 0); 
  int Find(void *buf, unsigned Bytes, unsigned Offset) const;
  int Find(const MemoryBuffer &buf, unsigned Offset = 0) const {
    return Find(buf.BufferPtr, buf.Len, Offset); 
  }
  
  // Deletes specifed number of bytes from starting position
  unsigned DeleteAt(unsigned Pos, unsigned Bytes);
  unsigned GetLength() const { return Len; }    // Returns logical length
  unsigned GetDimLen() const { return DimLen; } // Returns allocated length
  unsigned char *GetBufferPtr() { return BufferPtr; } 
  const unsigned char *GetBufferPtr() const { return BufferPtr; } 
  unsigned char *GetBuffer() { return BufferPtr; } 
  const unsigned char *GetBuffer() const { return BufferPtr; } 
  int Grow() { return Realloc(DimLen + GrowBy); } // Grow by default amount
  int ReAllocate(unsigned Bytes) { return Realloc(Bytes); } 
  int FreeExtra() { return Realloc(Len); }   // Frees unused space in buffer
  void ChgGrowByInc(int GB) { GrowBy = GB; } // Change GrowBy increment

  // Sets the logical length of the buffer to <= maximum length
  void SetLength(unsigned Bytes);

  // Returns true if the buffer references Null_Ptr
  int IsNull() const { return DataPtr == &MemoryRep::Null_Ptr; }
  int is_null() const { return DataPtr == &MemoryRep::Null_Ptr; }
  
private: 
  // Copies data from buf, resizing if Bytes > DimLen && GrowBy != 0
  void CopyN(const void *buf, unsigned Bytes); 

  // Inserts/replaces data pointed by s, up to Bytes starting at Pos
  unsigned InsReplAt(unsigned Pos, const void *buf,
		     unsigned Bytes, int Ins = 1);

  // Grows the string by the specified amount
  int Grow_By(unsigned Bytes) { return Realloc(DimLen + Bytes); }
  int EnsureUnique(); // Ensures that this string uniquely owns its data

  // Returns true if string references shared data or Null_Ptr
  int IsUnique() const { return DataPtr->RefCount == 1 || IsNull(); }

  // Check for string index being in bounds
  unsigned CheckIndex(unsigned i) {
    if(i >= Len) i = Len; // If not in bounds truncate to Len
    return i;
  }
  
private: // Memory allocation routines
  // If memory allocation for new buffer fails reference a null string of 0
  int Alloc(unsigned Bytes, int GB);

  // Redimenstions the buffer, returns 0 if memory not available
  int Realloc(unsigned NewDimLen, int Keep = 1);

  // Binds to the same data that s is bound to
  void Bind(const MemoryBuffer &buf);
  
  void UnBind(); // Unbinds this object from the shared buffer data

  // Unbinds this buffer and then binds it to data shared by "buf"
  void NewBinding(const MemoryBuffer &buf);

private: 
  MemoryRep *DataPtr;
  unsigned char *BufferPtr; // Pointer to start of buffer
  unsigned DimLen; // Maximum length without reallocation
  unsigned Len;    // Logical length of the buffer
  int GrowBy;      // Minimum resizing increment

public: // Convertion functions
  operator char *() const { return (char *)BufferPtr; }
  operator const char *() const { return (const char*)BufferPtr; }

public: // Overloaded operators
  unsigned char& operator[](unsigned i);
  void operator+=(const MemoryBuffer &buf) { Cat(buf.BufferPtr, buf.Len); }
  void operator+=(const unsigned char &byte) { Cat(byte); }
  void operator+=(const char &byte) { Cat(byte); }
  friend MemoryBuffer operator+(const MemoryBuffer &a, const MemoryBuffer &b);

  friend int operator==(const MemoryBuffer &a, const MemoryBuffer &b) {
    return BufferCompare(a, b) == 0;
  }

  friend int operator!=(const MemoryBuffer &a, const MemoryBuffer &b) {
    return BufferCompare(a, b) != 0;
  }

  friend int operator>(const MemoryBuffer &a, const MemoryBuffer &b) {
    return BufferCompare(a, b) > 0;
  }

  friend int operator>=(const MemoryBuffer &a, const MemoryBuffer &b) {
    return BufferCompare(a, b) >= 0;
  }
  
  friend int operator<(const MemoryBuffer &a, const MemoryBuffer &b) {
    return BufferCompare(a, b) < 0;
  }
  
  friend int operator<=(const MemoryBuffer &a, const MemoryBuffer &b) {
    return BufferCompare(a, b) <= 0;
  }

public: // Friend functions
  // Returns -1 if a < b, 0 if a == b, and 1 if a > b
  friend int BufferCompare(const MemoryBuffer &a, const MemoryBuffer &b);  
};

#endif  // __MEMBUF_HPP__
// ----------------------------------------------------------- // 
// ------------------------------- //
// --------- End of File --------- //
// ------------------------------- //

