// ------------------------------- //
// -------- Start of File -------- //
// ------------------------------- //
// ----------------------------------------------------------- //
// C++ Header File Name: vbscomm.h
// C++ Compiler Used: MSVC40, HP CPP 10.24, egcs-2.90.29 
// Produced By: Doug Gaer   
// File Creation Date: 09/20/1999
// Date Last Modified: 07/17/2000
// ----------------------------------------------------------- // 
// ---------- Include File Description and Details  ---------- // 
// ----------------------------------------------------------- // 
/*
The VBS C++ classes are copyright (c) 2000, by Douglas M. Gaer.
All those who put this code or its derivatives in a commercial
product MUST mention this copyright in their documentation for
users of the products in which this code or its derivative
classes are used. Otherwise, you have the freedom to redistribute
verbatim copies of this source code, adapt it to your specific
needs, or improve the code and release your improvements to the
public provided that the modified files carry prominent notices
stating that you changed the files and the date of any change.

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.
THE ENTIRE RISK OF THE QUALITY AND PERFORMANCE OF THIS SOFTWARE
IS WITH YOU. SHOULD ANY ELEMENT OF THIS SOFTWARE PROVE DEFECTIVE,
YOU WILL ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR, OR
CORRECTION.

The vbSerialComm class is a base class used to open a serial 
port for bidirectional communication. It includes several 
low-level functions needed by derived classes to initalize a 
serial port, transmit, and receive data.
*/
// ----------------------------------------------------------- //   
#ifndef __VBSCOMM_HPP__
#define __VBSCOMM_HPP__

// Include files
#if defined (__WIN32__)
#include <windows.h>
#elif defined (__UNIX__)
#include <unistd.h>  // UNIX standard function definitions 
#include <termios.h> // POSIX terminal control definitions 
#else
#error You must define a target platform: __WIN32__ or __UNIX__
#endif

// Type definitions
#if defined (__WIN32__)
typedef HANDLE scommDeviceHandle;
#elif defined (__UNIX__)
typedef int scommDeviceHandle;
#else
#error You must define a target platform: __WIN32__ or __UNIX__
#endif

// Serial communications base class
class vbSerialComm
{
public: // Enumerations
  enum {
    // Internal error codes used to report the last error
    scomm_NO_ERROR  = 0,      // No errors reported
    scomm_INVALID_ERROR_CODE, // Invalid error code

    scomm_BAUDRATE_ERROR,    // Invalid baud rate
    scomm_CS_ERROR,          // Invalid character size
    scomm_FLOWCONTROL_ERROR, // Invalid flow control
    scomm_INIT_ERROR,        // Cannot initialize serial device
    scomm_INVALIDPARM,       // Invalid initialization parameter
    scomm_OPEN_ERROR,        // Cannot open serial device
    scomm_PARITY_ERROR,      // Invalid parity
    scomm_RECEIVE_ERROR,     // Serial device receive error
    scomm_STOPBIT_ERROR,     // Invalid stop bit
    scomm_TRANSMIT_ERROR,    // Serial device transmit error

    // Exception codes added to handle variable block errors
    scomm_BLOCKACK_ERROR,    // Variable block acknowledgment error
    scomm_BLOCKHEADER_ERROR, // Bad variable block header
    scomm_BLOCKSIZE_ERROR,   // Bad variable block size
    scomm_BLOCKSYNC_ERROR    // Variable block synchronization error
  };
  
  enum {
    // Flow control constants
    scommHARD_FLOW,
    scommSOFT_FLOW,
    scommXON_XOFF,
    scommNO_FLOW_CONTROL,

    // Device access constants
    scommREAD_ONLY,
    scommWRITE_ONLY,
    scommREAD_WRITE
  };

public:
  vbSerialComm();
  virtual ~vbSerialComm();
  
public:
  int OpenSerialPort(char *device_name);
  int InitSerialPort();
  int InitSerialPort(char *device_name, int sp = 9600, char pr = 'N',
		     int cs = 8, int sb = 1,
		     int flow = vbSerialComm::scommNO_FLOW_CONTROL,
		     int bin_mode = 1);
  void Close();
  int RawRead(void *buf, int bytes);
  int RawWrite(const void *buf, int bytes);
  int Recv(void *buf, int bytes);
  int Send(const void *buf, int bytes);
  void SetBaudRate(int br) { baud_rate = br; }
  void SetCharacterSize(int cs) { character_size = cs; }
  void SetParity(char p) { parity = p; }
  void SetStopBits(int sb) { stop_bits = sb; }
  void SetFlowControl(int f) { flow_control = f; }
  scommDeviceHandle DeviceHandle() { return device_handle; }
  
  // NOTE: The Win32 API does not support non-binary mode transfers
  void BinaryMode() { binary_mode = 1; }
  void CharacterMode() { binary_mode = 0; }

  // Exception handling functions
  int GetSerialCommError() { return scomm_error; }
  int GetSerialCommError() const { return scomm_error; }
  void SetSerialCommError(int err) { scomm_error = err; }
  void ResetSerialCommError() { scomm_error = scomm_NO_ERROR; }
  void ResetError() { scomm_error = scomm_NO_ERROR; }
  const char *SerialCommExceptionMessage();
  
#if defined (__UNIX__)
  termios *GetTermIOS() { return &options; }
  int BytesRead() { return bytes_read; }
  int BytesMoved() { return bytes_moved; }
#endif

#if defined (__WIN32__)
  DCB *GetDeviceCB() { return &dcb; }
  DWORD BytesRead() { return bytes_read; }
  OVERLAPPED *Overlap() { return &ov; }
  DWORD BytesMoved() { return bytes_moved; }
  DWORD vbsGetLastError() { return last_err; }
#endif

protected:
  scommDeviceHandle device_handle; // Device handle for the port
  int baud_rate;      // Baud rate
  int character_size; // Character size
  char parity;        // Parity 
  int stop_bits;      // Stop bits
  int flow_control;   // Flow control
  int binary_mode;    // True to enable raw reads and raw writes
  int scomm_error;    // The last reported serial port error
  
protected: // POSIX terminal (serial) interface extensions
#if defined (__UNIX__)
  termios options; // Terminal control structure
  int bytes_read;  // Number of bytes read following a Read() call
  int bytes_moved; // Number of bytes written following a Write() call
#endif
  
protected: // WIN32 extensions
#if defined (__WIN32__)
  DCB dcb;           // Device-control block structure
  DWORD bytes_read;  // Address of number of bytes read following Read()
  OVERLAPPED ov;     // Pointer to structure needed for overlapped I/O  
  DWORD bytes_moved; // Pointer to number of bytes written following Write()
  DWORD last_err;    // Used to store the last error reported
#endif
};

#endif // __VBSCOMM_HPP__
// ----------------------------------------------------------- // 
// ------------------------------- //
// --------- End of File --------- //
// ------------------------------- //
